# DO NOT EDIT!
# This file was automatically generated by "generate_id_constants_from_stableid.py"

import enum

class BuffId(enum.Enum):
    NULL = 0
    RADAR25 = 1
    TAUNTB = 2
    DISABLEABILS = 3
    TRANSIENTMORPH = 4
    GRAVITONBEAM = 5
    GHOSTCLOAK = 6
    BANSHEECLOAK = 7
    POWERUSERWARPABLE = 8
    VORTEXBEHAVIORENEMY = 9
    CORRUPTION = 10
    QUEENSPAWNLARVATIMER = 11
    GHOSTHOLDFIRE = 12
    GHOSTHOLDFIREB = 13
    LEECH = 14
    LEECHDISABLEABILITIES = 15
    EMPDECLOAK = 16
    FUNGALGROWTH = 17
    GUARDIANSHIELD = 18
    SEEKERMISSILETIMEOUT = 19
    TIMEWARPPRODUCTION = 20
    ETHEREAL = 21
    NEURALPARASITE = 22
    NEURALPARASITEWAIT = 23
    STIMPACKMARAUDER = 24
    SUPPLYDROP = 25
    _250MMSTRIKECANNONS = 26
    STIMPACK = 27
    PSISTORM = 28
    CLOAKFIELDEFFECT = 29
    CHARGING = 30
    AIDANGERBUFF = 31
    VORTEXBEHAVIOR = 32
    SLOW = 33
    TEMPORALRIFTUNIT = 34
    SHEEPBUSY = 35
    CONTAMINATED = 36
    TIMESCALECONVERSIONBEHAVIOR = 37
    BLINDINGCLOUDSTRUCTURE = 38
    COLLAPSIBLEROCKTOWERCONJOINEDSEARCH = 39
    COLLAPSIBLEROCKTOWERRAMPDIAGONALCONJOINEDSEARCH = 40
    COLLAPSIBLETERRANTOWERCONJOINEDSEARCH = 41
    COLLAPSIBLETERRANTOWERRAMPDIAGONALCONJOINEDSEARCH = 42
    DIGESTERCREEPSPRAYVISION = 43
    INVULNERABILITYSHIELD = 44
    MINEDRONECOUNTDOWN = 45
    MOTHERSHIPSTASIS = 46
    MOTHERSHIPSTASISCASTER = 47
    MOTHERSHIPCOREENERGIZEVISUAL = 48
    ORACLEREVELATION = 49
    GHOSTSNIPEDOT = 50
    NEXUSPHASESHIFT = 51
    NEXUSINVULNERABILITY = 52
    ROUGHTERRAINSEARCH = 53
    ROUGHTERRAINSLOW = 54
    ORACLECLOAKFIELD = 55
    ORACLECLOAKFIELDEFFECT = 56
    SCRYERFRIENDLY = 57
    SPECTRESHIELD = 58
    VIPERCONSUMESTRUCTURE = 59
    RESTORESHIELDS = 60
    MERCENARYCYCLONEMISSILES = 61
    MERCENARYSENSORDISH = 62
    MERCENARYSHIELD = 63
    SCRYER = 64
    STUNROUNDINITIALBEHAVIOR = 65
    BUILDINGSHIELD = 66
    LASERSIGHT = 67
    PROTECTIVEBARRIER = 68
    CORRUPTORGROUNDATTACKDEBUFF = 69
    BATTLECRUISERANTIAIRDISABLE = 70
    BUILDINGSTASIS = 71
    STASIS = 72
    RESOURCESTUN = 73
    MAXIMUMTHRUST = 74
    CHARGEUP = 75
    CLOAKUNIT = 76
    NULLFIELD = 77
    RESCUE = 78
    BENIGN = 79
    LASERTARGETING = 80
    ENGAGE = 81
    CAPRESOURCE = 82
    BLINDINGCLOUD = 83
    DOOMDAMAGEDELAY = 84
    EYESTALK = 85
    BURROWCHARGE = 86
    HIDDEN = 87
    MINEDRONEDOT = 88
    MEDIVACSPEEDBOOST = 89
    EXTENDBRIDGEEXTENDINGBRIDGENEWIDE8OUT = 90
    EXTENDBRIDGEEXTENDINGBRIDGENWWIDE8OUT = 91
    EXTENDBRIDGEEXTENDINGBRIDGENEWIDE10OUT = 92
    EXTENDBRIDGEEXTENDINGBRIDGENWWIDE10OUT = 93
    EXTENDBRIDGEEXTENDINGBRIDGENEWIDE12OUT = 94
    EXTENDBRIDGEEXTENDINGBRIDGENWWIDE12OUT = 95
    PHASESHIELD = 96
    PURIFY = 97
    VOIDSIPHON = 98
    ORACLEWEAPON = 99
    ANTIAIRWEAPONSWITCHCOOLDOWN = 100
    ARBITERMPSTASISFIELD = 101
    IMMORTALOVERLOAD = 102
    CLOAKINGFIELDTARGETED = 103
    LIGHTNINGBOMB = 104
    ORACLEPHASESHIFT = 105
    RELEASEINTERCEPTORSCOOLDOWN = 106
    RELEASEINTERCEPTORSTIMEDLIFEWARNING = 107
    RELEASEINTERCEPTORSWANDERDELAY = 108
    RELEASEINTERCEPTORSBEACON = 109
    ARBITERMPCLOAKFIELDEFFECT = 110
    PURIFICATIONNOVA = 111
    CORRUPTIONBOMBDAMAGE = 112
    CORSAIRMPDISRUPTIONWEB = 113
    DISRUPTORPUSH = 114
    LIGHTOFAIUR = 115
    LOCKON = 116
    OVERCHARGE = 117
    OVERCHARGEDAMAGE = 118
    OVERCHARGESPEEDBOOST = 119
    SEEKERMISSILE = 120
    TEMPORALFIELD = 121
    VOIDRAYSWARMDAMAGEBOOST = 122
    VOIDMPIMMORTALREVIVESUPRESSED = 123
    DEVOURERMPACIDSPORES = 124
    DEFILERMPCONSUME = 125
    DEFILERMPDARKSWARM = 126
    DEFILERMPPLAGUE = 127
    QUEENMPENSNARE = 128
    ORACLESTASISTRAPTARGET = 129
    SELFREPAIR = 130
    AGGRESSIVEMUTATION = 131
    PARASITICBOMB = 132
    PARASITICBOMBUNITKU = 133
    PARASITICBOMBSECONDARYUNITSEARCH = 134
    ADEPTDEATHCHECK = 135
    LURKERHOLDFIRE = 136
    LURKERHOLDFIREB = 137
    TIMESTOPSTUN = 138
    SLAYNELEMENTALGRABSTUN = 139
    PURIFICATIONNOVAPOST = 140
    DISABLEINTERCEPTORS = 141
    BYPASSARMORDEBUFFONE = 142
    BYPASSARMORDEBUFFTWO = 143
    BYPASSARMORDEBUFFTHREE = 144
    CHANNELSNIPECOMBAT = 145
    TEMPESTDISRUPTIONBLASTSTUNBEHAVIOR = 146
    GRAVITONPRISON = 147
    INFESTORDISEASE = 148
    SS_LIGHTNINGPROJECTOR = 149
    PURIFIERPLANETCRACKERCHARGE = 150
    SPECTRECLOAKING = 151
    WRAITHCLOAK = 152
    PSYTROUSOXIDE = 153
    BANSHEECLOAKCROSSSPECTRUMDAMPENERS = 154
    SS_BATTLECRUISERHUNTERSEEKERTIMEOUT = 155
    SS_STRONGERENEMYBUFF = 156
    SS_TERRATRONARMMISSILETARGETCHECK = 157
    SS_MISSILETIMEOUT = 158
    SS_LEVIATHANBOMBCOLLISIONCHECK = 159
    SS_LEVIATHANBOMBEXPLODETIMER = 160
    SS_LEVIATHANBOMBMISSILETARGETCHECK = 161
    SS_TERRATRONCOLLISIONCHECK = 162
    SS_CARRIERBOSSCOLLISIONCHECK = 163
    SS_CORRUPTORMISSILETARGETCHECK = 164
    SS_INVULNERABLE = 165
    SS_LEVIATHANTENTACLEMISSILETARGETCHECK = 166
    SS_LEVIATHANTENTACLEMISSILETARGETCHECKINVERTED = 167
    SS_LEVIATHANTENTACLETARGETDEATHDELAY = 168
    SS_LEVIATHANTENTACLEMISSILESCANSWAPDELAY = 169
    SS_POWERUPDIAGONAL2 = 170
    SS_BATTLECRUISERCOLLISIONCHECK = 171
    SS_TERRATRONMISSILESPINNERMISSILELAUNCHER = 172
    SS_TERRATRONMISSILESPINNERCOLLISIONCHECK = 173
    SS_TERRATRONMISSILELAUNCHER = 174
    SS_BATTLECRUISERMISSILELAUNCHER = 175
    SS_TERRATRONSTUN = 176
    SS_VIKINGRESPAWN = 177
    SS_WRAITHCOLLISIONCHECK = 178
    SS_SCOURGEMISSILETARGETCHECK = 179
    SS_SCOURGEDEATH = 180
    SS_SWARMGUARDIANCOLLISIONCHECK = 181
    SS_FIGHTERBOMBMISSILEDEATH = 182
    SS_FIGHTERDRONEDAMAGERESPONSE = 183
    SS_INTERCEPTORCOLLISIONCHECK = 184
    SS_CARRIERCOLLISIONCHECK = 185
    SS_MISSILETARGETCHECKVIKINGDRONE = 186
    SS_MISSILETARGETCHECKVIKINGSTRONG1 = 187
    SS_MISSILETARGETCHECKVIKINGSTRONG2 = 188
    SS_POWERUPHEALTH1 = 189
    SS_POWERUPHEALTH2 = 190
    SS_POWERUPSTRONG = 191
    SS_POWERUPMORPHTOBOMB = 192
    SS_POWERUPMORPHTOHEALTH = 193
    SS_POWERUPMORPHTOSIDEMISSILES = 194
    SS_POWERUPMORPHTOSTRONGERMISSILES = 195
    SS_CORRUPTORCOLLISIONCHECK = 196
    SS_SCOUTCOLLISIONCHECK = 197
    SS_PHOENIXCOLLISIONCHECK = 198
    SS_SCOURGECOLLISIONCHECK = 199
    SS_LEVIATHANCOLLISIONCHECK = 200
    SS_SCIENCEVESSELCOLLISIONCHECK = 201
    SS_TERRATRONSAWCOLLISIONCHECK = 202
    SS_LIGHTNINGPROJECTORCOLLISIONCHECK = 203
    SHIFTDELAY = 204
    BIOSTASIS = 205
    PERSONALCLOAKINGFREE = 206
    EMPDRAIN = 207
    MINDBLASTSTUN = 208
    _330MMBARRAGECANNONS = 209
    VOODOOSHIELD = 210
    SPECTRECLOAKINGFREE = 211
    ULTRASONICPULSESTUN = 212
    IRRADIATE = 213
    NYDUSWORMLAVAINSTANTDEATH = 214
    PREDATORCLOAKING = 215
    PSIDISRUPTION = 216
    MINDCONTROL = 217
    QUEENKNOCKDOWN = 218
    SCIENCEVESSELCLOAKFIELD = 219
    SPORECANNONMISSILE = 220
    ARTANISTEMPORALRIFTUNIT = 221
    ARTANISCLOAKINGFIELDEFFECT = 222
    ARTANISVORTEXBEHAVIOR = 223
    INCAPACITATED = 224
    KARASSPSISTORM = 225
    DUTCHMARAUDERSLOW = 226
    JUMPSTOMPSTUN = 227
    JUMPSTOMPFSTUN = 228
    RAYNORMISSILETIMEDLIFE = 229
    PSIONICSHOCKWAVEHEIGHTANDSTUN = 230
    SHADOWCLONE = 231
    AUTOMATEDREPAIR = 232
    SLIMED = 233
    RAYNORTIMEBOMBMISSILE = 234
    RAYNORTIMEBOMBUNIT = 235
    TYCHUSCOMMANDOSTIMPACK = 236
    VIRALPLASMA = 237
    NAPALM = 238
    BURSTCAPACITORSDAMAGEBUFF = 239
    COLONYINFESTATION = 240
    DOMINATION = 241
    EMPBURST = 242
    HYBRIDCZERGYROOTS = 243
    HYBRIDFZERGYROOTS = 244
    LOCKDOWNB = 245
    SPECTRELOCKDOWNB = 246
    VOODOOLOCKDOWN = 247
    ZERATULSTUN = 248
    BUILDINGSCARAB = 249
    VORTEXBEHAVIORERADICATOR = 250
    GHOSTBLAST = 251
    HEROICBUFF03 = 252
    CANNONRADAR = 253
    SS_MISSILETARGETCHECKVIKING = 254
    SS_MISSILETARGETCHECK = 255
    SS_MAXSPEED = 256
    SS_MAXACCELERATION = 257
    SS_POWERUPDIAGONAL1 = 258
    WATER = 259
    DEFENSIVEMATRIX = 260
    TESTATTRIBUTE = 261
    TESTVETERANCY = 262
    SHREDDERSWARMDAMAGEAPPLY = 263
    CORRUPTORINFESTING = 264
    MERCGROUNDDROPDELAY = 265
    MERCGROUNDDROP = 266
    MERCAIRDROPDELAY = 267
    SPECTREHOLDFIRE = 268
    SPECTREHOLDFIREB = 269
    ITEMGRAVITYBOMBS = 270
    CARRYMINERALFIELDMINERALS = 271
    CARRYHIGHYIELDMINERALFIELDMINERALS = 272
    CARRYHARVESTABLEVESPENEGEYSERGAS = 273
    CARRYHARVESTABLEVESPENEGEYSERGASPROTOSS = 274
    CARRYHARVESTABLEVESPENEGEYSERGASZERG = 275
    PERMANENTLYCLOAKED = 276
    RAVENSCRAMBLERMISSILE = 277
    RAVENSHREDDERMISSILETIMEOUT = 278
    RAVENSHREDDERMISSILETINT = 279
    RAVENSHREDDERMISSILEARMORREDUCTION = 280
    CHRONOBOOSTENERGYCOST = 281
    NEXUSSHIELDRECHARGEONPYLONBEHAVIOR = 282
    NEXUSSHIELDRECHARGEONPYLONBEHAVIORSECONDARYONTARGET = 283
    INFESTORENSNARE = 284
    INFESTORENSNAREMAKEPRECURSORREHEIGHTSOURCE = 285
    NEXUSSHIELDOVERCHARGE = 286
    PARASITICBOMBDELAYTIMEDLIFE = 287
    TRANSFUSION = 288
    INHIBITORZONETEMPORALFIELD = 289

for item in BuffId:
    assert not item.name in globals()
    globals()[item.name] = item
